@@ -4,7 +4,7 @@ import google.api_core.exceptions
 from google.cloud.pubsub_v1 import SubscriberClient
 from google.protobuf.duration_pb2 import Duration
 from google.protobuf.field_mask_pb2 import FieldMask
-from google.pubsub_v1.types.pubsub import ExpirationPolicy, RetryPolicy, Subscription as _Subscription
+from google.pubsub_v1.types.pubsub import ExpirationPolicy, PushConfig, RetryPolicy, Subscription as _Subscription
 
 
 logger = logging.getLogger(__name__)
@@ -28,6 +28,7 @@ class Subscription:
     :param int|None expiration_time: number of seconds after which the subscription is deleted (infinite time if None)
     :param float minimum_retry_backoff: minimum number of seconds after the acknowledgement deadline has passed to exponentially retry delivering a message to the subscription
     :param float maximum_retry_backoff: maximum number of seconds after the acknowledgement deadline has passed to exponentially retry delivering a message to the subscription
+    :param str|None push_endpoint: if this is a push subscription, this is the URL to which messages should be pushed; leave as `None` if this is a pull subscription
     :return None:
     """
 
@@ -43,6 +44,7 @@ class Subscription:
         expiration_time=THIRTY_ONE_DAYS,
         minimum_retry_backoff=10,
         maximum_retry_backoff=600,
+        push_endpoint=None,
     ):
         if namespace and not name.startswith(namespace):
             self.name = f"{namespace}.{name}"
@@ -67,6 +69,24 @@ class Subscription:
             maximum_backoff=Duration(seconds=maximum_retry_backoff),
         )
 
+        self.push_endpoint = push_endpoint
+
+    @property
+    def is_pull_subscription(self):
+        """Return `True` if this is a pull subscription.
+
+        :return bool:
+        """
+        return self.push_endpoint is None
+
+    @property
+    def is_push_subscription(self):
+        """Return `True` if this is a push subscription.
+
+        :return bool:
+        """
+        return self.push_endpoint is not None
+
     def __repr__(self):
         """Represent the subscription as a string.
 
@@ -137,24 +157,30 @@ class Subscription:
         """
         return f"projects/{project_name}/subscriptions/{subscription_name}"
 
+    def _create_proto_message_subscription(self):
+        """Create a Proto message subscription from the instance to be sent to the Pub/Sub API.
+
+        :return google.pubsub_v1.types.pubsub.Subscription:
+        """
+        if self.push_endpoint:
+            push_config = {"push_config": PushConfig(mapping=None, push_endpoint=self.push_endpoint)}
+        else:
+            push_config = {}
+
+        return _Subscription(
+            mapping=None,
+            name=self.path,  # noqa
+            topic=self.topic.path,
+            ack_deadline_seconds=self.ack_deadline,  # noqa
+            message_retention_duration=self.message_retention_duration,  # noqa
+            expiration_policy=self.expiration_policy,  # noqa
+            retry_policy=self.retry_policy,  # noqa
+            **push_config,
+        )
+
     def _log_creation(self):
         """Log the creation of the subscription.
 
         :return None:
         """
         logger.debug("Subscription %r created.", self.path)
-
-    def _create_proto_message_subscription(self):
-        """Create a Proto message subscription from the instance to be sent to the Pub/Sub API.
-
-        :return google.pubsub_v1.types.pubsub.Subscription:
-        """
-        return _Subscription(
-            mapping=None,
-            name=self.path,
-            topic=self.topic.path,
-            ack_deadline_seconds=self.ack_deadline,  # noqa
-            message_retention_duration=self.message_retention_duration,  # noqa
-            expiration_policy=self.expiration_policy,  # noqa
-            retry_policy=self.retry_policy,  # noqa
-        )
